/*******************************************************************
  User space IRQ handling emulator for i386 Linux based PCs
  It uses VM86_PLUS services developed for DOSEMU

  emu_irq.c

  (C) Copyright 1998 by Pavel Pisa 

  This code is distributed under the Gnu General Public Licence. 
  See file COPYING for details.
*******************************************************************/
 
//#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#include <sys/io.h>
#include <errno.h>

#include <asm/vm86.h>
#include <sys/syscall.h>

#include "emu_irq.h"

#define NEW_SYS_vm86  166

static inline int vm86_plus(int function, int param)
{
	int __res;
	__asm__ __volatile__("int $0x80\n"
                             :"=a" (__res):"a" ((int)NEW_SYS_vm86), "b" (function), "c" (param));
	return __res;
} 

#define EMU_IRQ_MAX_INTNO 15

int emu_irq_thread_prio;
int emu_irq_initialized=0;
int emu_irq_terminate=0;
int emu_irq_global_disable=0;
pthread_t emu_irq_thread=(pthread_t)0;
pthread_mutex_t emu_irq_mutex=PTHREAD_MUTEX_INITIALIZER;

#define EMU_IRQ_REQUEST_OP_REQUEST 1
#define EMU_IRQ_REQUEST_OP_RELEASE 2

int emu_irq_request_operation;
int emu_irq_request_intno;
int emu_irq_request_result;
sem_t emu_irq_request_sem;
pthread_mutex_t emu_irq_request_mutex=PTHREAD_MUTEX_INITIALIZER;

struct emu_irq 
{
        emu_irq_handler *handler;
        void   *dev_id;
	int    enabled;
	int    pending;
} emu_irq_arr[EMU_IRQ_MAX_INTNO+1] = {{0,},};

void emu_irq_cli(void)
{
	pthread_mutex_lock(&emu_irq_mutex);
	emu_irq_global_disable++;
	pthread_mutex_unlock(&emu_irq_mutex);
}

void emu_irq_sti(void)
{
	pthread_mutex_lock(&emu_irq_mutex);
	if(--emu_irq_global_disable<=0)
		if(emu_irq_thread)
			pthread_kill(emu_irq_thread, EMU_IRQ_SIG);
	pthread_mutex_unlock(&emu_irq_mutex);
}

/* Empty function, never called, signal is permanently blocked */
void emu_irq_hanle_sig(int signo)
{
}

/* All signal processing is carried synchronously in the thread */
void *emu_irq_thread_task(void *arg)
{
        unsigned emu_irq_set;
        int intno;
        int ret;
	int signo;
	sigset_t wait_sigset;

	sigemptyset(&wait_sigset);
	sigaddset(&wait_sigset,EMU_IRQ_SIG);
	sigprocmask(SIG_BLOCK,&wait_sigset,NULL);

	while(!emu_irq_terminate){
		ret = sigwait(&wait_sigset, &signo);

		if(emu_irq_request_operation){
			intno = emu_irq_request_intno;
			if(emu_irq_request_operation==EMU_IRQ_REQUEST_OP_REQUEST){
				ret = vm86_plus(VM86_REQUEST_IRQ,(EMU_IRQ_SIG<<8)|intno);
				if (ret==intno) {
					emu_irq_arr[intno].enabled=1;
				}else{
					emu_irq_arr[intno].handler = NULL;
					ret = -EPERM;
				}
			}else if(emu_irq_request_operation==EMU_IRQ_REQUEST_OP_RELEASE){
			        ret = vm86_plus(VM86_FREE_IRQ,intno);
			        emu_irq_arr[intno].handler = NULL;
			        emu_irq_arr[intno].enabled = 0;
			}else{
				ret = -1;
			}
			emu_irq_request_result = ret;
			emu_irq_request_operation = 0;
			sem_post(&emu_irq_request_sem);
		}

		if(emu_irq_global_disable>0)
			continue;

		emu_irq_set = vm86_plus(VM86_GET_IRQ_BITS,0);
		for(intno = 0; emu_irq_set; emu_irq_set>>=1, intno++) {
			if(emu_irq_set & 1)
				emu_irq_arr[intno].pending=1;
			if(emu_irq_arr[intno].pending&&emu_irq_arr[intno].handler&&
			   (emu_irq_arr[intno].enabled>0)) {
				emu_irq_handler *handler;
				void *dev_id;
				pthread_mutex_lock(&emu_irq_mutex);
				handler=emu_irq_arr[intno].handler;
				dev_id=emu_irq_arr[intno].dev_id;
				if(!handler||(emu_irq_arr[intno].enabled<=0)){
					pthread_mutex_unlock(&emu_irq_mutex);
					continue;
				}
				emu_irq_arr[intno].pending=0;
				ret = vm86_plus(VM86_GET_AND_RESET_IRQ, intno);
				handler(intno,dev_id,NULL);
				pthread_mutex_unlock(&emu_irq_mutex);
			}
		}
	}
	return NULL;
}

int emu_irq_thread_prio_set(int prio)
{
	int ret=0;

	if((prio>sched_get_priority_max(SCHED_FIFO))||
	   (prio<sched_get_priority_min(SCHED_FIFO)))
		return EINVAL;

	emu_irq_thread_prio=prio;

	pthread_mutex_lock(&emu_irq_mutex);
	if(emu_irq_thread && !emu_irq_terminate){
		struct sched_param sched_par;
		sched_par.sched_priority = prio;
		ret=pthread_setschedparam(emu_irq_thread, SCHED_FIFO, &sched_par);
	}
	pthread_mutex_unlock(&emu_irq_mutex);

	return 	ret>0?-ret:0;
}


void release_all_emu_irq(void);

int emu_irq_initialize(void)
{
	pthread_mutex_lock(&emu_irq_mutex);
	if(!emu_irq_initialized){
		struct sigaction emu_irq_sig;
		struct sigaction old_sig;
		sigset_t wait_sigset;
		pthread_attr_t th_attr;
		struct sched_param sched_par;

		emu_irq_initialized = 1;

		sem_init(&emu_irq_request_sem, 0, 0);

		/* Signal has to be blocked in all threads to allow process it through sigwait */
		sigemptyset(&wait_sigset);
		sigaddset(&wait_sigset,EMU_IRQ_SIG);
		sigprocmask(SIG_BLOCK,&wait_sigset,NULL);
		pthread_sigmask(SIG_BLOCK,&wait_sigset,NULL);

		/* Set something sensible, but generally ignored */
		memset(&emu_irq_sig,0,sizeof(emu_irq_sig));
		emu_irq_sig.sa_handler = emu_irq_hanle_sig;
		sigemptyset(&emu_irq_sig.sa_mask);
		emu_irq_sig.sa_flags = SA_RESTART;
		emu_irq_sig.sa_restorer = 0;
		sigaction(EMU_IRQ_SIG, &emu_irq_sig, &old_sig);

		/* Ensure restoration on exit */
		atexit(&release_all_emu_irq);

		if(!emu_irq_thread_prio){
			emu_irq_thread_prio=sched_get_priority_max(SCHED_FIFO);
			emu_irq_thread_prio+=sched_get_priority_min(SCHED_FIFO);
			emu_irq_thread_prio/=2;
		}
		pthread_attr_init(&th_attr);
		pthread_attr_setschedpolicy(&th_attr,SCHED_FIFO);
		sched_par.sched_priority = emu_irq_thread_prio;
		pthread_attr_setschedparam(&th_attr,&sched_par);
		if(pthread_create(&emu_irq_thread, &th_attr, emu_irq_thread_task, NULL)){
			emu_irq_initialized = 0;
			pthread_mutex_unlock(&emu_irq_mutex);
			errno = EBUSY;
			return -EBUSY;
		}
	}

	pthread_mutex_unlock(&emu_irq_mutex);
	return 0;
}

int request_emu_irq(unsigned int intno, emu_irq_handler *handler,
                    unsigned long flags, const char *device, void *dev_id)
{
	int ret;

	if(!emu_irq_initialized) {
		ret=emu_irq_initialize();
		if(ret)
			return ret;
        }

        if(intno > EMU_IRQ_MAX_INTNO) { errno = EINVAL; return -EINVAL; }

	pthread_mutex_lock(&emu_irq_request_mutex);
	pthread_mutex_lock(&emu_irq_mutex);
        if(emu_irq_arr[intno].handler){
		errno = EBUSY;
		pthread_mutex_unlock(&emu_irq_mutex);
		return -EBUSY;
	}
        emu_irq_arr[intno].handler = handler;
        emu_irq_arr[intno].dev_id = dev_id;

	emu_irq_request_result = -1;
	emu_irq_request_intno = intno;

	pthread_mutex_unlock(&emu_irq_mutex);

	emu_irq_request_operation = EMU_IRQ_REQUEST_OP_REQUEST;

	pthread_kill(emu_irq_thread, EMU_IRQ_SIG);

	sem_wait(&emu_irq_request_sem);

	ret = emu_irq_request_result;
	if(ret<0)
		errno = -ret;

	pthread_mutex_unlock(&emu_irq_request_mutex);
	return ret;
}

int  release_emu_irq(int intno, void *dev_id)
{ 
        int ret;
	pthread_mutex_lock(&emu_irq_request_mutex);
	pthread_mutex_lock(&emu_irq_mutex);

	emu_irq_request_result = -1;
	emu_irq_request_intno = intno;

	pthread_mutex_unlock(&emu_irq_mutex);

	emu_irq_request_operation = EMU_IRQ_REQUEST_OP_RELEASE;

	pthread_kill(emu_irq_thread, EMU_IRQ_SIG);

	sem_wait(&emu_irq_request_sem);

	ret = emu_irq_request_result;
	if(ret<0)
		errno = -ret;

	pthread_mutex_unlock(&emu_irq_request_mutex);
        return ret;
}

void release_all_emu_irq(void)
{
        int intno;
	int join_fl=0;

	emu_irq_terminate=1;

	pthread_mutex_lock(&emu_irq_mutex);
	if(emu_irq_initialized){
		if(emu_irq_thread){
			pthread_kill(emu_irq_thread, EMU_IRQ_SIG);
			join_fl=1;
		}
	}
	pthread_mutex_unlock(&emu_irq_mutex);

	if(join_fl){
		pthread_join(emu_irq_thread,NULL);
	}

        if(emu_irq_initialized) {
                for(intno=0; intno<=EMU_IRQ_MAX_INTNO; intno++)
                        if(emu_irq_arr[intno].handler != NULL) 
                                release_emu_irq(intno,emu_irq_arr[intno].dev_id);
                emu_irq_initialized = 0;
        }
}
